/**
 * Copyright (C) 2010 Cloudfarming <info@cloudfarming.nl>
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.geoviewer.jxmap.map;

import com.vividsolutions.jts.geom.Geometry;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import nl.cloudfarming.client.geoviewer.LayerObject;
import nl.cloudfarming.client.geoviewer.Palette;
import nl.cloudfarming.client.geoviewer.jxmap.render.GeometryRenderer;
import nl.cloudfarming.client.geoviewer.jxmap.render.GeometryRendererFactory;
import nl.cloudfarming.client.util.GeometryUtil;
import org.jdesktop.swingx.JXMapViewer;
import org.jdesktop.swingx.mapviewer.GeoPosition;
import org.netbeans.api.visual.model.ObjectScene;
import org.netbeans.api.visual.widget.Widget;
import org.openide.nodes.Node;

/**
 * An ItemWidget is a object on the map which is used for interaction from
 * the user with a dynamic object on the map (a {@link LayerItem}).
 * Someone can see a tooltip for a dynamic object and can use a context menu
 * (right mouse click on an object).
 * 
 * @author Maarten Hollenga
 */
public class LayerObjectWidget extends Widget {

    private static final int ICON_SIZE = 20;
    private String id;
    private final JXMapViewer mapViewer;
    private final LayerObject layerObject;
    private final Palette palette;

    public void setId(String id) {
        this.id = id;
    }

    public String getId() {
        return id;
    }

    /**
     * Initialize the widget for a dynamic object on the specified location
     * with a tooltip and a context menu.
     * 
     * @param scene A class that keeps associations between widgets and the 
     * objects that model them
     * @param layerItem The dynamic object on the map
     * @param tooltiptext The text to display at the tooltip
     * @param explorerManager node explorer manager
     */
    public LayerObjectWidget(final ObjectScene scene, final LayerObject layerObject, Palette palette,  JXMapViewer mapViewer) {
        super(scene);
        this.palette = palette;
        this.mapViewer = mapViewer;
        this.layerObject = layerObject;
        this.id = layerObject.getKey();
        setPreferredSize(new Dimension(ICON_SIZE, ICON_SIZE));
        setToolTipText(layerObject.getTooltipText());
        setForeground(this.palette.getColorForValue(layerObject.getKeyAttributeValue()));

    }

    @Override
    protected void paintWidget() {
        //
        // get the renderer for the geometry in the layer object and render the geometry
        //
        GeometryRenderer renderer = GeometryRendererFactory.getRenderer(layerObject.getGeometry());
        getGraphics().setColor(getForeground());

        getGraphics().setColor(this.palette.getColorForValue(layerObject.getKeyAttributeValue()));

        Rectangle clientArea = renderer.paint(layerObject.getGeometry(), this.mapViewer, getGraphics(), getState().isSelected());
        clientArea = (clientArea != null) ? clientArea : super.calculateClientArea();

        setPreferredBounds(clientArea);
    }

    @Override
    public boolean isHitAt(Point localLocation) {
        if (isVisible() && getBounds().contains(localLocation)) {

            Geometry geo = layerObject.getGeometry();
            // TODO: put in util
            Rectangle viewportBounds = mapViewer.getViewportBounds();

            int mapX = (int) (viewportBounds.getX() + localLocation.getX());

            int mapY = (int) (viewportBounds.getY() + localLocation.getY());
            Point2D newPoint = new Point(mapX, mapY);

            GeoPosition pos = mapViewer.getTileFactory().pixelToGeo(newPoint, mapViewer.getZoom());

            com.vividsolutions.jts.geom.Point pnt = GeometryUtil.getPoint(pos.getLongitude(), pos.getLatitude());
            return geo.intersects(pnt);

        } else {
            return false;
        }
    }
  
}
