/**
 * Copyright (C) 2010 Cloudfarming <info@cloudfarming.nl>
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.geoviewer.jxmap.map;

import java.awt.BorderLayout;
import java.awt.EventQueue;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.InputEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;
import javax.swing.JComponent;
import javax.swing.JPopupMenu;
import javax.swing.SwingUtilities;
import nl.cloudfarming.client.geoviewer.Layer;
import nl.cloudfarming.client.geoviewer.LayerNode;
import nl.cloudfarming.client.geoviewer.LayerObject;
import nl.cloudfarming.client.geoviewer.LayerObjectNode;
import nl.cloudfarming.client.geoviewer.MapController;
import nl.cloudfarming.client.geoviewer.ObjectLayer;
import nl.cloudfarming.client.model.Bean;
import nl.cloudfarming.client.model.BeanNotFoundException;
import org.jdesktop.swingx.JXMapViewer;
import org.netbeans.api.visual.action.ActionFactory;
import org.netbeans.api.visual.action.PopupMenuProvider;
import org.netbeans.api.visual.action.WidgetAction.WidgetMouseEvent;
import org.netbeans.api.visual.model.ObjectScene;
import org.netbeans.api.visual.model.ObjectSceneEvent;
import org.netbeans.api.visual.model.ObjectSceneEventType;
import org.netbeans.api.visual.model.ObjectSceneListener;
import org.netbeans.api.visual.model.ObjectState;
import org.netbeans.api.visual.widget.LayerWidget;
import org.netbeans.api.visual.widget.Widget;
import org.openide.explorer.ExplorerManager;
import org.openide.nodes.Node;
import org.openide.nodes.NodeEvent;
import org.openide.nodes.NodeListener;
import org.openide.nodes.NodeMemberEvent;
import org.openide.nodes.NodeReorderEvent;
import org.openide.util.Exceptions;
import org.openide.util.Utilities;

/**
 * Layer panel containint objects
 * object layers will be painted as images with a backgroundPainter by default
 * when the layer is activated, the actual objects (widgets) will be painted over the background.
 *
 * @author Timon Veenstra
 */
public class ObjectLayerPanel extends LayerPanel implements ObjectSceneListener {

    private static final Logger LOGGER = Logger.getAnonymousLogger();
    final ObjectScene scene = new ObjectScene();
    private final JComponent view = scene.createView();
    private boolean incomingSelectionChange = false;
    private boolean outgoingSelectionChange = false;
    private boolean removingObjects = false;
    //
    // widget layer
    private final LayerWidget mainLayer = new LayerWidget(scene);

    public ObjectLayerPanel(final MapController controller, Node node, Layer layer, JXMapViewer mapViewer, int id) {
        super(controller, layer, mapViewer, id, RootMapPanel.LayerGroup.LAYER_ACTIVE);
        assert (layer instanceof ObjectLayer);
        final ObjectLayer objectLayer = (ObjectLayer) layer;


        setLayout(new BorderLayout());
        this.scene.setOpaque(false);
        this.scene.addChild(this.mainLayer);
        this.scene.setMaximumBounds(new Rectangle(0, 0, Integer.MAX_VALUE, Integer.MAX_VALUE));
        this.scene.addObjectSceneListener(this, ObjectSceneEventType.OBJECT_STATE_CHANGED, ObjectSceneEventType.OBJECT_SELECTION_CHANGED);

        node.addNodeListener(new NodeListener() {

            @Override
            public void childrenAdded(NodeMemberEvent ev) {
                final Node[] addedObjects = new Node[ev.getDelta().length];
                int i = 0;
                for (Node node : ev.getDelta()) {
                    if (node instanceof LayerObjectNode) {
                        addedObjects[i] = node;
                        i++;
                    }
                }
                addObjects(addedObjects);
            }

            @Override
            public void childrenRemoved(NodeMemberEvent ev) {
                final List<Node> removeObjects = new ArrayList<Node>();
                for (Node node : ev.getDelta()) {
                    if (node instanceof LayerObjectNode) {
                        objectLayer.removeObject(((LayerObjectNode) node).getLayerObject());
                        removeObjects.add(node);
                    }
                }
                removeObjects(removeObjects);
                //view.repaint();
            }

            @Override
            public void childrenReordered(NodeReorderEvent ev) {
            }

            @Override
            public void nodeDestroyed(NodeEvent ev) {
            }

            @Override
            public void propertyChange(PropertyChangeEvent evt) {
            }
        });

        if (node.getChildren() == null || node.getChildren().getNodesCount() == 0) {
            LOGGER.warning("Empty ObjectLayer: The LayerNode did not contain any children.");
        } else {
            addObjects(node.getChildren().getNodes());
        }

        add(view);
        setActive(true);
    }

    public static boolean isRightMouseButton(WidgetMouseEvent anEvent) {
        return ((anEvent.getModifiers() & InputEvent.BUTTON3_MASK) == InputEvent.BUTTON3_MASK);
    }

    @Override
    public void propertyChange(PropertyChangeEvent evt) {
        if (!outgoingSelectionChange && ExplorerManager.PROP_SELECTED_NODES.equals(evt.getPropertyName())) {
            Node[] newSelection = (Node[]) evt.getNewValue();
            Set<Node> selection = new HashSet<Node>();

            for (Node selected : newSelection) {
                if (selected instanceof LayerObjectNode) {
                    // only add to selection if the object is part of this scene
                    if (scene.findStoredObject(selected) != null) {
                        selection.add(selected);
                    }
                }
            }

            incomingSelectionChange = true;
            scene.setSelectedObjects(selection);
            incomingSelectionChange = false;
            view.repaint();

        }
    }

    /**
     * remove objects
     * if remove is called from the event dispatch thread it ill be
     * executed immediate, if not it will be scheduled on the dispatch thread.
     *
     * @param objects
     */
    private void removeObjects(final List<Node> objects) {
        if (EventQueue.isDispatchThread()) {
            internalRemoveObjects(objects);
        } else {
            EventQueue.invokeLater(new Runnable() {

                @Override
                public void run() {
                    internalRemoveObjects(objects);
                }
            });
        }
    }

    private void internalRemoveObjects(List<Node> objects) {
        assert objects != null;
        removingObjects = true;
        for (Node node : objects) {
            List<Widget> widgets = this.scene.findWidgets(node);
            if (widgets != null) {
                for (Widget w : widgets) {
                    mainLayer.removeChild(w);
                }
                this.scene.removeObject(node);
            }

        }
        this.scene.validate();
        removingObjects = false;
    }

    /**
     * add objects
     * if called from the event dispatch thread it ill be
     * executed immediate, if not it will be scheduled on the dispatch thread.
     *
     * @param objects
     */
    private void addObjects(final Node[] nodes) {
        if (SwingUtilities.isEventDispatchThread()) {
            internalAddObjects(nodes);
        } else {
            SwingUtilities.invokeLater(new Runnable() {

                @Override
                public void run() {
                    internalAddObjects(nodes);
                }
            });
        }
    }

    /**
     * Direct add
     *
     * @param nodes
     */
    private void internalAddObjects(final Node[] nodes) {
        assert SwingUtilities.isEventDispatchThread();

        for (Node node : nodes) {
            addObject(node);
        }
        scene.validate();
    }

    /**
     * Add an object to the panel.
     * this method will add the object to the scene and create a widget attached to it.
     *
     *
     * @param objectNode
     */
    private void addObject(final Node node) {
        assert SwingUtilities.isEventDispatchThread();
        LayerObject layerObject = node.getLookup().lookup(LayerObject.class);
        if (layerObject == null || this.scene.getObjects().contains(node)) {
            return;
        }
        LayerObjectWidget widget = new LayerObjectWidget(this.scene, layerObject, getLayer().getPalette(), getMapViewer());


        widget.getActions().addAction(scene.createObjectHoverAction());
        widget.getActions().addAction(scene.createSelectAction());

        widget.getActions().addAction(ActionFactory.createPopupMenuAction(new PopupMenuProvider() {

            @Override
            public JPopupMenu getPopupMenu(Widget widget, Point localLocation) {
                JPopupMenu menu = Utilities.actionsToPopup(node.getActions(true), node.getLookup());
                return menu;
            }
        }));

        this.mainLayer.addChild(widget);
        this.scene.addObject(node, widget);

    }

    /**
     * activate or deactivate this layer
     * when layer gets activated, the view will be made visible, painting all containing objects as widgets
     * and providing interaction.
     *
     * @param active
     */
    void setActive(boolean active) {
        this.view.setVisible(active);
        setBackgroundPainter(active ? null : super.getPainter());
        this.view.repaint();
    }

    @Override
    public void objectAdded(ObjectSceneEvent event, Object addedObject) {
    }

    @Override
    public void objectRemoved(ObjectSceneEvent event, Object removedObject) {
    }

    @Override
    public void objectStateChanged(ObjectSceneEvent event, Object changedObject, ObjectState previousState, ObjectState newState) {
        if (!removingObjects && !incomingSelectionChange && changedObject instanceof Node) {
            Node node = (Node) changedObject;

            outgoingSelectionChange = true;
            //
            // add node to the selection
            //
            if (!previousState.isSelected() && newState.isSelected()) {
                try {
                    getLayer().getDataProvider().addBeanToSelection(node.getLookup().lookup(Bean.class));
                } catch (BeanNotFoundException ex) {
                    Exceptions.printStackTrace(ex);
                }
            }
            //
            // remove node from the selection if it is not removed
            //
            if (previousState.isSelected() && !newState.isSelected()) {
                try {
                    getLayer().getDataProvider().removeBeanFromSelection(node.getLookup().lookup(Bean.class));
                } catch (BeanNotFoundException ex) {
                    Exceptions.printStackTrace(ex);
                }
            }
            outgoingSelectionChange = false;
        }
    }

    @Override
    public void selectionChanged(ObjectSceneEvent event, Set<Object> previousSelection, Set<Object> newSelection) {
        for (Object o : previousSelection) {
            Widget widget = event.getObjectScene().findWidget(o);
            widget.setForeground(getLayer().getPalette().getColorForState(ObjectState.createNormal().deriveSelected(false)));
        }
        for (Object o : newSelection) {
            Widget widget = event.getObjectScene().findWidget(o);
            widget.setForeground(getLayer().getPalette().getColorForState(ObjectState.createNormal().deriveSelected(true)));
        }
    }

    @Override
    public void highlightingChanged(ObjectSceneEvent event, Set<Object> previousHighlighting, Set<Object> newHighlighting) {
    }

    @Override
    public void hoverChanged(ObjectSceneEvent event, Object previousHoveredObject, Object newHoveredObject) {
    }

    @Override
    public void focusChanged(ObjectSceneEvent event, Object previousFocusedObject, Object newFocusedObject) {
    }

    @Override
    public void onMapMouseEvent(MapMouseEvent mapMouseEvent) {
        view.dispatchEvent(mapMouseEvent.getEvent());
        if (mapMouseEvent.getEvent().isPopupTrigger() && isLayerObjectHit(mapMouseEvent.getEvent().getPoint(), scene)) {
            mapMouseEvent.consume();
        }

    }

    @Override
    public void mapMousePressed(MapMouseEvent mapMouseEvent) {
    }

    @Override
    public void mapMouseReleased(MapMouseEvent mapMouseEvent) {
    }

    @Override
    public void mapMouseDragged(MapMouseEvent mapMouseEvent) {
    }

    /**
     * Recursive method checking if a layerobject is clicked
     */
    private boolean isLayerObjectHit(final Point point, final Widget widget) {
        if ((widget instanceof LayerObjectWidget) && widget.isHitAt(widget.convertSceneToLocal(point))) {
            return true;
        }

        for (final Widget child : widget.getChildren()) {
            if (isLayerObjectHit(point, child)) {
                return true;
            }
        }

        return false;
    }
}
